/* This file is part of mailfrom filter.
   Copyright (C) 2005, Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <resolv.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include <pthread.h>

#include "mailfrom.h"

struct mx_buffer {
	unsigned pref;
	char *name;
};

static int
comp_pref(const void *a, const void *b)
{
	const struct mx_buffer *ma = a, *mb = b;
	if (ma->pref > mb->pref)
		return 1;
	else if (ma->pref < mb->pref)
		return -1;
	return 0;
}

/* Obtain MX records for domain name HOST. Return them in mxbuf, sorted by
   preference in ascending order.
   Notice that we *have* to use old BIND-4-style method, since glibc folks
   in their incredible wiseness have not exported new interface functions
   from libresolve.so. The only way to access them would be to link with
   -lresolve statically.
   
   Let's hope that someday this stupidity will change. */
   
static getmx_status
_getmx(char *host, char *answer, size_t answer_size, mxbuf_t mxbuf)
{
	int i, n, nmx;
	struct mx_buffer mx_buffer[MAXMXCOUNT];
	HEADER *hp;
	unsigned char *eom, *cp;
	unsigned short qdcount, ancount;
	
	debug1(100,"Getting MX records for %s", host);
	n = res_query (host, C_IN, T_MX, answer, answer_size);
	if (n < 0) {
		debug2(10, "res_query failed (errno=%s, h_errno=%d)",
			   mu_strerror(errno), h_errno);
		switch (h_errno) {
		case NO_DATA:
		case NO_RECOVERY:
		case HOST_NOT_FOUND:
			return getmx_not_found;

		case TRY_AGAIN:
		case -1:
			return getmx_temp_failure;

		default:
			mu_error("res_nsearch(%s) failed with unexpected h_errno %d",
				 host, h_errno);
			return getmx_failure;
		}
	}

	if (n > answer_size)
		n = answer_size;

	hp = (HEADER*)answer;
	cp = (unsigned char *) answer + HFIXEDSZ;
	eom = (unsigned char *) answer + n;
	
	/* Skip query part */
	for (qdcount = ntohs((unsigned short)hp->qdcount);
	     qdcount--;
	     cp += n + QFIXEDSZ)
	{
		if ((n = dn_skipname(cp, eom)) < 0)
			return getmx_failure;
	}

	ancount = ntohs((unsigned short)hp->ancount);

	/* Collect MX records */
	for (i = nmx = 0; i < ancount; i++) {
		unsigned short pref, type;
		char tname[NS_MAXDNAME];

		if ((n = dn_expand((unsigned char *)answer,
				   eom, cp, tname, sizeof tname)) < 0)
			break;
		cp += n;
		GETSHORT(type, cp);
		cp += INT16SZ + INT32SZ;
		GETSHORT(n, cp);
		if (type != T_MX) {
			debug2(100,"unexpected answer type %d, size %d\n",
				   type, n);
			cp += n;
			continue;
		}
		GETSHORT(pref, cp);
		if ((n = dn_expand((u_char *)answer, eom, cp,
				   tname, sizeof tname)) < 0)
			break;
		cp += n;
		mx_buffer[nmx].pref = pref;
		mx_buffer[nmx].name = strdup(tname);
		debug2(100,"MX %u %s", mx_buffer[nmx].pref,
			   mx_buffer[nmx].name);
		if (nmx++ > MAXMXCOUNT)
			break;
	}

	/* Sort according to preference value */
	qsort(mx_buffer, nmx, sizeof mx_buffer[0], comp_pref);

	/* Prepare return value */
	memset(mxbuf, 0, sizeof(mxbuf_t));
	for (i = 0; i < nmx; i++) 
		mxbuf[i] = mx_buffer[i].name;
	return getmx_success;
}

static int
is_ipaddr(char *str)
{
	for (; *str; str++)
		if (!(isdigit(*str) || *str == '.'))
			return 0;
	return 1;
}

static pthread_mutex_t getmx_mutex = PTHREAD_MUTEX_INITIALIZER;

getmx_status
getmx(char *host, mxbuf_t mxbuf)
{
	char hbuf[NI_MAXHOST];
	getmx_status status = getmx_failure;

	pthread_mutex_lock(&getmx_mutex);
	if (is_ipaddr(host)) {
		struct sockaddr_in sa;
		memset(&sa, 0, sizeof sa);
		sa.sin_family = AF_INET;
		if (inet_aton(host, &sa.sin_addr)
		    && getnameinfo((struct sockaddr *)&sa, sizeof sa,
				   hbuf, sizeof(hbuf),
				   NULL, 0,
				   0) == 0
		    && !is_ipaddr(hbuf))
			host = hbuf;
		else
			host = NULL;
	}
	
	if (host) {		
		unsigned char *answer = malloc(MAXPACKET);
		if (!answer)
			status = getmx_failure;
		else {
			char *p;

			for (p = host; p; ) {
				status = _getmx(p, answer, MAXPACKET, mxbuf);
				if (status == getmx_success)
					break;
				p = strchr(p, '.');
				if (!p)
					break;
				p++;
			}
			free(answer);
		}
	}
	pthread_mutex_unlock(&getmx_mutex);
	return status;
}
		
